'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _shockedCommon = require('shocked-common');

var _TrackerClient = require('./TrackerClient');

var _TrackerClient2 = _interopRequireDefault(_TrackerClient);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var EventEmitter = require('events');

function createClient(host) {
  var WebSocket = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : global.WebSocket;

  if (!host.startsWith('ws://') && !host.startsWith('wss://')) {
    throw new Error('Invalid host ' + host + '. Host should start with ws:// or wss://');
  }

  // Using an array, since the number of trackers is not expected
  // to be very high, typically 2 trackers at a time
  var trackers = [];
  function findTracker(trackerId) {
    return trackers.find(function (tracker) {
      return tracker.group === trackerId;
    });
  }

  var parser = (0, _shockedCommon.createParser)();

  var eventManager = new EventEmitter();

  var reconnectTimerHandle = null;
  function setupReconnection(interval) {
    if (reconnectTimerHandle) {
      return;
    }

    reconnectTimerHandle = setTimeout(function () {
      reconnectTimerHandle = null;
      // eslint-disable-next-line no-use-before-define
      client.reconnect();
    }, interval);
  }

  function clearRetry() {
    if (reconnectTimerHandle) {
      clearTimeout(reconnectTimerHandle);
      reconnectTimerHandle = null;
    }
  }

  function connection(remoteUrl) {
    if (remoteUrl === null) {
      return null;
    }

    var sock = new WebSocket(remoteUrl);
    sock.onerror = function () {
      // console.error('Error', e);
    };

    sock.onopen = function () {
      // Let all the trackers know tha we are now connected
      trackers.forEach(function (tracker) {
        // eslint-disable-next-line no-use-before-define
        tracker.onConnect(client);
      });

      // Trigger the connect event
      eventManager.emit('connect');
    };

    sock.onclose = function (e) {
      if (e.code !== 1000 && e.code !== 1005) {
        // Try to reconnect again after sometime
        setupReconnection(1000);
      }

      trackers.forEach(function (tracker) {
        // Let all the trackers know that the client is not available
        tracker.onDisconnect();
      });

      // Fire the close event on client
      eventManager.emit('disconnect');
    };

    sock.onmessage = function (e) {
      parser.parse(e.data);
    };

    return sock;
  }

  parser.onTrackerCreateNew = function (trackerId, serial, data, apis) {
    var tracker = findTracker(trackerId);
    if (tracker) {
      tracker.onCreate(serial, data, apis);
    }
  };

  parser.onTrackerCreateUpdate = function (trackerId, serial, actions) {
    var tracker = findTracker(trackerId);
    if (tracker) {
      tracker.onUpdate(serial, actions);
    }
  };

  parser.onTrackerAction = function (trackerId, action) {
    var tracker = findTracker(trackerId);
    if (tracker) {
      tracker.onAction(action);
    }
  };

  parser.onTrackerApiResponse = function (trackerId, apiId, status, response, params) {
    var tracker = findTracker(trackerId);
    if (tracker) {
      tracker.onApiResponse(apiId, status, response, params);
    }
  };

  parser.onTrackerEvent = function (trackerId, event, data) {
    var tracker = findTracker(trackerId);
    if (tracker) {
      tracker.emit(event, data);
    }
  };

  // Initialize with a connection attempt
  var socket = null;
  var url = null;

  var client = {
    on: function on(event, listener) {
      if (listener) {
        eventManager.on(event, listener);
      }
    },

    off: function off(event, listener) {
      if (listener) {
        eventManager.removeListener(event, listener);
      }
    },

    isConnected: function isConnected() {
      return socket && socket.readyState === WebSocket.OPEN;
    },

    connect: function connect(path) {
      url = '' + host + path;
      if (client.isConnected() && socket.url === url) {
        return true;
      }

      if (socket !== null) {
        socket.close();
      }

      socket = connection(url);
      return true;
    },

    reconnect: function reconnect() {
      if (url === null) {
        // No prior url to reconnect to
        return false;
      }

      // Since its a reconnect attempt, we will close existing socket
      if (socket !== null) {
        socket.close();
      }

      socket = connection(url);
      return true;
    },

    close: function close() {
      if (socket) {
        socket.close();
      }
      socket = null;

      clearRetry();
    },

    send: function send(data) {
      if (!client.isConnected()) {
        return;
      }

      socket.send(data);
    },

    createTracker: function createTracker(trackerId, store) {
      var params = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};

      // make sure this trackerId is unique for this client
      if (trackers.find(function (tracker) {
        return tracker.trackerId === trackerId;
      })) {
        throw new Error('A tracker for ' + trackerId + ' already exists on the client. There can only be one tracker for one trackerId.');
      }

      var tracker = new _TrackerClient2.default(store, trackerId, params, function () {
        var idx = trackers.indexOf(tracker);
        if (idx >= 0) {
          trackers.splice(idx, 1);
        }
      });

      // Include the tracker in the list
      trackers.push(tracker);

      // Send a connect event if the client is already connect
      if (client.isConnected()) {
        tracker.onConnect(client);
      }

      return tracker;
    }
  };

  return client;
}

exports.default = createClient;